
import os,profile,sys
from stat import *
from tempfile import mkstemp
from sys import argv
from struct import pack,unpack
from collections import deque
from array import array
from bisect import bisect

#import psyco
#psyco.full()

SIZE = 32*1024
BLOCKSIZE = 0x4000
BLOCKS = SIZE / BLOCKSIZE
header = "SM2\x1a"

def checksum(str):
	sum = 0
	for i in range(128):
		sum += unpack("<I", str[i*128:i*128+4])[0]
	return (sum & 0xffffffff)

def romtrunc(contents, size):
	i = size - 1
	repeat = contents[i]
	if ord(repeat) == 0x00 or ord(repeat) == 0xff:
		i -= 1
		while i > -1 and contents[i] == repeat:
			i -= 1
	return contents[:i + 2]

# read a whole file
def readfile(name):
	fd = open(name, "rb")
	contents = fd.read()
	fd.close()
	return os.stat(name)[ST_SIZE], contents

# write a whole file
def writefile(name, contents):
	fd = open(name, "wb")
	fd.write(contents)
	fd.close()

def writetemp(contents):
	(fd, name) = mkstemp()
	os.write(fd,contents)
	os.close(fd)
	return name

def mapperdetect(rom):
	r0 = 0
	if len(rom) <= 0x8000:
		return 0
	r3 = r4 = r5 = 0
	r6 = 0x8000
	r7 = -1
	while r6:
		r1 = rom[r0]
		r0 += 1
		if r1 == 0x32:
			r1 = r2 = rom[r0]
			r0 += 1
			r1 = r1 | (r2<<8)
			if r1 == 0x4000 or r1 == 0x8000:
				r3 += 1
				r0 += 2
			if r1 == 0xa000:
				r4 += 1
				r0 += 2
			if r1 == (r7 >> 16):
				r5 += 1
				r0 += 2
		r6 -= 1
	r1 = r5 + 1
	if r3 > r1:
		return 1
	if r4 > r1:
		return 2
	return 0

class apack:
	def compress(self, z):
		'''the actual apack compressor

		   returns compressed string'''
		inname = writetemp(z)
		(tmp, outname) = mkstemp()
		os.close(tmp)
		path = "c:/documents and settings/michael/my documents/gba/pocketnes_compy/apack.exe"
		os.spawnl(os.P_WAIT, path, "apack.exe", inname, outname)
		os.remove(inname)
		(size, str) = readfile(outname)
		os.remove(outname)
		return str

def gbh(filein, fileout):
	size, contents = readfile(filein)
	if size % BLOCKSIZE:
		size += (BLOCKSIZE - (size % BLOCKSIZE))
		contents = contents.ljust(size, "\0")
	mapper = mapperdetect(contents)
	print mapper
	chksum = checksum(contents)
	blocks_cmp = []
	blocks = []
	diff = []
	compressed = []
	for i in xrange(0,size,BLOCKSIZE):
		chunk = contents[i:i+BLOCKSIZE]
		trunc = romtrunc(chunk, BLOCKSIZE)
		cmp = apack().compress(trunc)
		diff.append(len(trunc)-len(cmp))
		blocks_cmp.append(cmp)
		blocks.append(trunc)
	spaceleft = SIZE
	while spaceleft:
		largest = -1
		index = -1
		for j in xrange(len(diff)):
			if len(blocks[j]) <= spaceleft and diff[j] > largest:
				largest = diff[j]
				index = j
		if largest < 0:
			break
		print largest, index
		spaceleft -= len(blocks[index])
		blocks[index] = blocks_cmp[index]
		compressed.append(index)
		diff[index] = -1
	newfile = header + pack("<III", chksum, size, mapper)
	offset = 16 + size/BLOCKSIZE * 4
	rest = ""
	for i in xrange(len(blocks)):
		if i in compressed:
			newfile += pack("<I", offset | 0x80000000)
		else:
			newfile += pack("<I", offset)
		offset += len(blocks[i])
		rest += blocks[i]
	newfile += rest
	writefile(fileout, newfile)

if __name__ == "__main__":
	if len(argv) == 3:
		gbh(argv[1], argv[2])
		#profile.run('npc(argv[1], argv[2])')
